#import "ImageDownloaderBlockingCache.h"
#import "ImageDownloaderBlockingCacheContainer.h"
#import "ImageDownloaderBlockingPatch.h"
#import "QCGLImage-ImageDownloaderBlockingCacheUtilities.h"


static NSMutableDictionary *ImageDownloaderBlockingCachePool=nil;

@implementation ImageDownloaderBlockingCache

// each glContext needs to have an independent image cache, so this factory creates (or reissues) a unique cache for each glContext.
+ (ImageDownloaderBlockingCache *)cacheForContext:(QCOpenGLContext *)glContext
{
	if( !ImageDownloaderBlockingCachePool )
	{
//		NSLog(@"ImageDownloaderBlockingCache::cacheForContext(%08x)  Allocating pool.",glContext);
		ImageDownloaderBlockingCachePool = [[NSMutableDictionary alloc] init];
	}

	// we use [glContext description] as the key --- it looks something like this:
	//		<QCOpenGLContext = 0x15ED7160 | renderer = ATI Radeon X1600 OpenGL Engine (2.0 ATI-1.4.56)>
	// and the 0x15ED7160 part becomes the unique identifier
	ImageDownloaderBlockingCache *ourCache=[ImageDownloaderBlockingCachePool objectForKey:[glContext description]];
	if( !ourCache )
	{
//		NSLog(@"ImageDownloaderBlockingCache::cacheForContext(%08x)  Cache doesn't exist for this glContext; allocating new cache.",glContext);
		ourCache = [[ImageDownloaderBlockingCache alloc] initWithContext:glContext];
		[ImageDownloaderBlockingCachePool setObject:ourCache forKey:[glContext description]];
	}
	else
	{
//		NSLog(@"ImageDownloaderBlockingCache::cacheForContext(%08x)  Reusing existing cache.",glContext);
	}

	return ourCache;
}

+ (unsigned long int)totalCacheMemory
{
	unsigned long int t=0;

	NSEnumerator *enumerator=[ImageDownloaderBlockingCachePool keyEnumerator];
	id key;
	while( (key=[enumerator nextObject]) )
	{
		t += [[ImageDownloaderBlockingCachePool objectForKey:key] cacheMemory];
	}
	return t;
}

+ (void)cleanAllCaches
{
//	NSLog(@"ImageDownloaderBlockingCache::cleanAllCaches()  before totalCacheMemory=%lu",[ImageDownloaderBlockingCache totalCacheMemory]);

	NSEnumerator *enumerator=[ImageDownloaderBlockingCachePool keyEnumerator];
	id key;
	while( (key=[enumerator nextObject]) )
	{
		[[ImageDownloaderBlockingCachePool objectForKey:key] cleanCache];
	}

//	NSLog(@"ImageDownloaderBlockingCache::cleanAllCaches()  after  totalCacheMemory=%lu",[ImageDownloaderBlockingCache totalCacheMemory]);
}

+ (float)cacheVRAMLimitPercent
{
	static float ImageDownloaderBlockingCacheVRAMLimitPercent=-1.0;

	if( ImageDownloaderBlockingCacheVRAMLimitPercent == -1.0 )
	{
		NSDictionary *d=[[NSBundle bundleForClass:[self class]] infoDictionary];
		ImageDownloaderBlockingCacheVRAMLimitPercent=[[d objectForKey:@"ImageDownloaderBlockingCacheVRAMLimitPercent"] floatValue];
//		NSLog(@"ImageDownloaderBlockingCacheVRAMLimitPercent = %f",ImageDownloaderBlockingCacheVRAMLimitPercent);
	}
	return ImageDownloaderBlockingCacheVRAMLimitPercent;
}













- (id)initWithContext:(QCOpenGLContext *)context
{
//	NSLog(@"ImageDownloaderBlockingCache::initWithContext();");
	id z=[super init];
	if(z)
	{
		_context=context;
		_cache=[[NSMutableDictionary alloc] init];
		_cacheMemory=0;
	}
	return z;
}

- (void)removeFromCache:(NSURL *)url
{
	ImageDownloaderBlockingCacheContainer *c=[_cache objectForKey:url];
	_cacheMemory -= [c estimatedSize];
//	NSLog(@"ImageDownloaderBlockingCache::removeFromCache()  hits = %4lu, url = %@",[c hits],url);
	QCGLCIImage *q=[c image];
	[q release];
	[_cache removeObjectForKey:url];
}

- (void)clearCache
{
//	NSLog(@"ImageDownloaderBlockingCache::clearCache();");

	NSEnumerator *enumerator=[_cache keyEnumerator];
	id key;
	while( (key=[enumerator nextObject]) )
	{
		[self removeFromCache:key];
	}
}

- (void)cleanCache
{
//	NSLog(@"ImageDownloaderBlockingCache::cleanCache();");

	unsigned int killed=0;

	// remove those with no cache hits
	{
		NSEnumerator *enumerator=[_cache keyEnumerator];
		id key;
		while( (key=[enumerator nextObject]) )
		{
			ImageDownloaderBlockingCacheContainer *c=[_cache objectForKey:key];
			if( [c hits] == 0 )
			{
				[self removeFromCache:key];
				++killed;
			}
		}
	}
	
	// if nothing has been removed already, remove the first one
	if( !killed )
	{
		NSEnumerator *enumerator=[_cache keyEnumerator];
		id key = [enumerator nextObject];
		if( key )
			[self removeFromCache:key];
	}
}

- (QCGLCIImage *) requestImageUsingCache:(NSURL *)url
{
	QCGLCIImage *qcglOut;

	if( [_cache objectForKey:url] )
	{
//		NSLog(@"ImageDownloaderBlockingCache::requestImageUsingCache(%@)  Retrieving from cache.",url);
		ImageDownloaderBlockingCacheContainer *c=[_cache objectForKey:url];
		qcglOut = [c image];
	}
	else
	{
//		NSLog(@"ImageDownloaderBlockingCache::requestImageUsingCache(%@)  Not in cache; performing NSURLRequest.",url);

		NSURLRequest *request=
			[NSURLRequest requestWithURL:url cachePolicy:NSURLRequestReloadIgnoringCacheData timeoutInterval:0];
		NSAssert(request,@"ImageDownloaderBlockingCache::requestImageUsingCache()  request == 0");

		NSError *error=nil;
		NSURLResponse *response;
		NSData *data=
			[NSURLConnection sendSynchronousRequest:request returningResponse:&response error:&error];

		if(![data length])
		{
//			NSLog(@"ImageDownloaderBlockingCache::requestImageUsingCache()  NSData: len %08x",[data length]);
//			NSLog(@"ImageDownloaderBlockingCache::requestImageUsingCache()  NSError: %08x",error);

			NSBundle *thisBundle=[NSBundle bundleForClass:[self class]];
			NSString *path=[thisBundle pathForResource:@"bad" ofType:@"png"];
			if(!path)
			{
				NSLog(@"ImageDownloaderBlockingCache::requestImageUsingCache()  can't find bad.png");
				return NO;
			}
			NSURL *errorURL=[NSURL fileURLWithPath:path];
			data=[NSData dataWithContentsOfURL:errorURL options:0 error:&error];
		}


		{
			NSAssert(data,@"ImageDownloaderBlockingCache::requestImageUsingCache()  data == 0");
			CIImage *image=[CIImage imageWithData:data];
			NSAssert(image,@"ImageDownloaderBlockingCache::requestImageUsingCache()  image == 0");

			qcglOut=[
						[QCGLCIImage alloc]
						initWithCIImage:[image retain]
						options:0
					];
			NSAssert(qcglOut,@"ImageDownloaderBlockingCache::requestImageUsingCache()  qcglOut == 0");

			ImageDownloaderBlockingCacheContainer *c=[[ImageDownloaderBlockingCacheContainer alloc] initWithImage:qcglOut];

//			NSLog(@"ImageDownloaderBlockingCache::requestImageUsingCache(%@)  estimated size: %lu",url,[c estimatedSize]);

			if( ([ImageDownloaderBlockingCache totalCacheMemory] + [c estimatedSize]) <
					([_context textureMemorySize] * [ImageDownloaderBlockingCache cacheVRAMLimitPercent]) )
			{
				// it fits in the amount of vram we're allowed to use

//				NSLog(@"it fits!");
			}
			else
			{
				// it doesn't fit in the amount of vram we're allowed to use

//				NSLog(@"it doesn't fit.");
				[ImageDownloaderBlockingCache cleanAllCaches];
			}


			// add to cache
			{
				[_cache setValue:c forKey:(id)[url retain]];
				_cacheMemory += [c estimatedSize];
//				NSLog(@"ImageDownloaderBlockingCache::cacheMemory=%lu",[self cacheMemory]);
//				NSLog(@"ImageDownloaderBlockingCache::totalCacheMemory=%lu",[ImageDownloaderBlockingCache totalCacheMemory]);
			}

			[image release];
		}
	}

	return qcglOut;
}

- (unsigned long int)cacheMemory
{
	return _cacheMemory;
}
@end